/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	SWIACTX.CPP: Primary module
//
#pragma warning ( disable : 4786 )      	//  overlength debug info 
#include <crtdbg.h>
#include <assert.h>
#include "swiactx.h"
#include "olehelp.h"


#define HRFAIL(hr)  return bHrFail(hr)
#define E_PROLOG_FAIL  (OLE_S_LAST+1)

DECL_IP_H(ISupportErrorInfo);
DECL_IP_H(IErrorInfo);

extern "C"
{
	install_t install();
	install_t uninstall();
};



///////////////////////////////////////////////////////////////////////
//      The ActiveX "Journal" (registration) object/
///////////////////////////////////////////////////////////////////////

ACTXJOURNAL actxJournal;
static bool g_bInstalled = false;
static bool g_bDispatchErrorsAsExceptions = false;


///////////////////////////////////////////////////////////////////////
//      Common atoms and functor handles
///////////////////////////////////////////////////////////////////////

atom_t	atomActxError;  		//  "activeXError"

//  Boolean atoms
atom_t	atomTrue;       		//  'true'
atom_t	atomFalse;      		//  'false'

//  Function invocation type atoms
atom_t	atomFunc;       		//  'func'
atom_t	atomPropGet;    		//  'propget'
atom_t	atomPropPut;    		//  'propput'
atom_t	atomPropPutRef; 		//  'propputref'

//  Function argument atoms
atom_t	atomOptional;   		//  'optional'

//  Generated functors
functor_t	fnctrActxObject;		//  "activeXObject"
functor_t	fnctrActxError; 		//  "activeXError"
functor_t	fnctrActxSafearray;		//  "activeXSafearray"
functor_t	fnctrActxVariant;		//	"activeXVariant"

//  System Local ID
static LCID g_lcid;

//
//	Initialize all the constants used between this module and the prolog engine
//
void init_constants()
{ 
	g_lcid = ::GetSystemDefaultLCID();

	atomTrue			= PL_new_atom("true");
	atomFalse      		= PL_new_atom("false");
	atomFunc			= PL_new_atom("func");
	atomOptional    	= PL_new_atom("optional");
	atomPropGet    		= PL_new_atom("propget");
	atomPropPut    		= PL_new_atom("propput");
	atomPropPutRef  	= PL_new_atom("propputref");
	atomActxError   	= PL_new_atom("activeXError");

	//  Error functor, activeXError/3: 
	//		activeXError("NameOfFunction","ErrorDescription",TermCausingProblem).
	fnctrActxError  	= PL_new_functor( atomActxError, 3 );
	//  COM Object reference wrapper functor, activeXObject/1:  
	//		activeXObject(IndexOfSmartPointerInMap).
	fnctrActxObject 	= PL_new_functor( PL_new_atom("activeXObject"), 1 );
	//	Safearray wrapper functor, activeXSafearray:
	//		activeXSafearray(VariantType,ListOfElements).
	fnctrActxSafearray  = PL_new_functor( PL_new_atom("activeXSafearray"), 2 );
	//  VARIANT wrapper for unusual types
	fnctrActxVariant	= PL_new_functor( PL_new_atom("activeXVariant"), 2 );
}

//
//      If the term given is a "special" atom, return its enum, else ENONE
//      
EKEYWD eKeywordCheck ( term_t t )
{
	atom_t atm;
	if ( ! PL_get_atom( t, & atm ) )
		return ENONE;
	if ( atm == atomTrue )
		return ETRUE;
	else
	if ( atm == atomFalse )
		return EFALSE;
	else
	if ( atm == atomOptional )
		return EOPTIONAL;
	return ENONE;
}


//
//	Convert the invocation atom ('func', 'propget', etc.) to
//		the correct INVOKE_KIND enum.
//
DWORD invokeAtom ( term_t tKind )
{
	atom_t atm;
	if ( ! PL_get_atom( tKind, & atm ) )
		return 0;
	if ( atm == atomFunc )
		return INVOKE_FUNC;
	else
	if ( atm == atomPropGet )
		return INVOKE_PROPERTYGET;
	else
	if ( atm == atomPropPut )
		return INVOKE_PROPERTYPUT;
	else
	if ( atm == atomPropPutRef )
		return INVOKE_PROPERTYPUTREF;
	return 0;
}


//  Debugging point for Prolog exception returns
foreign_t raisePrologException ( term_t tExcept )
{
	return PL_raise_exception( tExcept );
}

//  Debugging point for Prolog failure
bool bHrFail ( HRESULT hr )
{
	PL_fail;
}

//
//	Create an instance using the COM "smart pointer" routines.
//	Input is either a Prog ID or class GUID.
//
HRESULT hrCreateInstance ( 
	IP_IUnknown & ipUnk, 
	TSZC tszcProgIdOrClassId, 
	DWORD dwClsContext )
{
	ZSTRW zswProgId( tszcProgIdOrClassId );
	return ipUnk.CreateInstance( (LPOLESTR) zswProgId.Tszc(), NULL, dwClsContext );
}


//
//      Convert the the Prolog list of arguments according to the
//      ITypeInfo given.  Store the results into the output parameter array.
//
HRESULT hrConvertArgList ( 
	IP_ITypeInfo & ipInfo,
	BINDPTR & bindPtr,
	term_t tArgList,
	ACTXDPARAMS & params, 
	ZSTR & zsError )
{
	int cParams = bindPtr.lpfuncdesc->cParams;
	params.resize( cParams );

	term_t thead = PL_new_term_ref();
	term_t tlist = PL_copy_term_ref( tArgList );

	for ( int iparam = 0; 
		  iparam < cParams && PL_get_list( tlist, thead, tlist );  
		  iparam++ )
	{
		TYPEDESC & tdesc = bindPtr.lpfuncdesc->lprgelemdescParam[iparam].tdesc;

		//      Determine the variant type of the required argument
		VARTYPE vt = vtGetVarType( ipInfo, tdesc );
		if ( vt == VT_ILLEGAL )
		{
			zsError.Format( "variant %d type handling not implemented",  iparam );
			return E_FAIL;
		}

		if ( ! bConvertTermToVariant( vt, thead, params.VVars()[cParams - 1 - iparam] ) )
		{
			zsError.Format( "unable to convert argument %d to vartype %d", 
							iparam, 
							vt );
			return E_FAIL;
		}
	}       

	if ( PL_get_nil( tlist ) )
		return S_OK;

	zsError = "argument list size incorrect";
	return E_FAIL;
}

//
//	Get the method invocation information from the IDispatch reference
//
HRESULT hrGetInvokeInfo ( 
	IP_IDispatch & ipId,    		//  [in]	IDispatch interface ref
	TSZC tszcMethod,				//  [in]	Name of method
	DWORD invkind,  				//  [in]	Type of invocation (INVOKEKIND)
	MEMBERID & mbrId,       		//  [out]   Member id
	IP_ITypeInfo & ipInfo,  		//  [out]   Interface type info
	BINDPTR & bindPtr )     		//  [out]   Function binding information
{
	IP_ITypeComp ipComp;
	DESCKIND descKind;
	IP_ITypeInfo ipTiBind;
	ZSTRW zswMethod( tszcMethod );
	OLECHAR * wszMethod = (OLECHAR *) zswMethod.Tszc();
	DWORD dwHash = ::LHashValOfNameSys( SYS_WIN32, g_lcid, wszMethod );
	HRESULT hr = S_OK;
	do
	{
		hr = ipId->GetTypeInfo( 0, 0, & ipInfo );
		if ( FAILED(hr) )
			break;

		hr = ipInfo->GetTypeComp( & ipComp );
		if ( FAILED(hr) )
			break;

		hr = ipComp->Bind( wszMethod,
							dwHash, 
							invkind,
							& ipTiBind, 
							& descKind, 
							& bindPtr );
		if ( FAILED(hr) )
			break;

		mbrId = bindPtr.lpfuncdesc->memid;
	} while (false);

	return hr;      
}

//
//	Same as above when ITypeInfo pointer and BINDPTR are not required.
//
HRESULT hrGetInvokeInfo ( 
	IP_IDispatch & ipId,    		//  [in]	IDispatch interface ref
	TSZC tszcMethod,				//  [in]	Name of method
	DWORD invkind,  				//  [in]	Type of invocation
	MEMBERID & mbrId )      		//  [out]	Member id
{
	IP_ITypeInfo ipInfo;
	BINDPTR bindPtr;
	return hrGetInvokeInfo( ipId, tszcMethod, invkind, mbrId, ipInfo, bindPtr );
}


//
//	Create a COM object and its activeXObject wrapper functor 
//
foreign_t
actxCreateObject ( term_t tProgId, term_t tIDispatchFunctor )
{
	term_t tArgInvalid = 0;

	//  Validate the progid/guid argument
	int itype = PL_term_type( tProgId );
	if ( itype != PL_ATOM )
		tArgInvalid = tProgId;
	else
	if ( PL_term_type( tIDispatchFunctor ) != PL_VARIABLE )
		tArgInvalid = tIDispatchFunctor;

	//	Raise an exception if invalid
	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxCreateObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	char * tszcProgId;
	if ( ! PL_get_atom_chars( tProgId, & tszcProgId ) )
		HRFAIL( E_PROLOG_FAIL );

	//  Create the object
	IP_IUnknown ipUnk;
	HRESULT hr = hrCreateInstance( ipUnk, tszcProgId );
	if ( FAILED(hr) )
	{
		ZSTR zsHr;
		zsHr.Format("0x%x", hr);
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxCreateObject",
						PL_CHARS, "CoCreateInstance error",
						PL_CHARS, zsHr.Tszc() );
		return raisePrologException( tExcept );
	}

	//  Query the IDispatch interface from the object.
	IP_IDispatch ipDispatch( ipUnk );
	if ( ! ipDispatch )
	{
		ZSTR zsHr;
		zsHr.Format("0x%x", hr);
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxCreateObject",
						PL_CHARS, "Not automation interface (IDispatch unavailable)",
						PL_CHARS, zsHr.Tszc() );
		return raisePrologException( tExcept );
	}

	//  Register the object and unify the result
	ACTXTOKEN atkn = actxJournal.registerObject( ipDispatch );
	if ( ! PL_unify_term( tIDispatchFunctor,
						PL_FUNCTOR, fnctrActxObject, 
						PL_INTEGER, atkn ) )
		PL_fail;

	PL_succeed;
}


//
//	Release the activeXObject.
//
foreign_t actxReleaseObject ( term_t tIDispatchFunctor )
{
	IP_IDispatch ipId;
	ACTXTOKEN tknObj;

	if ( ! bValidActiveXReference( tIDispatchFunctor, ipId, & tknObj ) )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxReleaseObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tIDispatchFunctor );
		return raisePrologException( tExcept );
	}
	actxJournal.erase( tknObj );

	PL_succeed;
}

//
//	Allow a currently registered IDispatch interface to be QI'd for
//		another interface.  If successful, the new interface is 
//      QI'd for its IDispatch.  If successful, that is returned.
//
//   In other words, only IDispatchs are ever registered or returned to Prolog.
//
foreign_t actxQueryInterface ( 
	term_t tIDispatchFunctor,
	term_t tGuid,
	term_t tResult )
{
	term_t tArgInvalid = 0;
	IP_IDispatch ipId;
	GUID guid;

	if (  ! bValidActiveXReference( tIDispatchFunctor, ipId ) )
		tArgInvalid = tIDispatchFunctor;
	else
	if ( ! PL_is_variable( tResult ) )
		tArgInvalid = tResult;

	if ( ! tArgInvalid )
	{
		char * pch = NULL;
		if ( ! (    PL_get_chars( tGuid, & pch, CVT_ALL | BUF_DISCARDABLE )
			     && bGuidFromString( pch, guid ) ) )
		{
			tArgInvalid = tGuid;
		}
	}

	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxInvokeObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	IP_IDispatch ipIdNew;
	IP_IUnknown ipUnk;
	HRESULT hr = ipId->QueryInterface( guid, (void **) & ipUnk );

	if ( SUCCEEDED(hr) )
	{
		//  If we got that interface, see if it converts to IDispatch.
		ipIdNew = ipUnk;
		//  Cannot get IDispatch from new interface-- it's useless to us
		if ( ! ipIdNew )
			HRFAIL( RPC_S_INTERFACE_NOT_FOUND );

		ACTXTOKEN atkn = actxJournal.registerObject( ipIdNew );
		PL_unify_term( tResult,
						PL_FUNCTOR, fnctrActxObject, 
						PL_INTEGER, atkn );

		PL_succeed;
	}

	HRFAIL(hr);
}

//
//	Parse the invocation term.  It may be a single atom (the method name)
//   	or it may be a list containing the method name and the invocation type
//     	atom ('func', 'propget', etc.).
//
bool bParseInvocationList ( 
	term_t tInvokeList,     	//  [in]    list term
	DWORD & dwInvoke,       	//  [out]   OLE invocation flags
	TSZC & tszcMethod )     	//  [out]   Method name string
{
	//  Default the value of the invocation flags
	dwInvoke = INVOKE_FUNC;

	//  Allow the argument to be a single atom: the method name
	if ( PL_is_atom( tInvokeList ) )
	{
		char * tszAtom;
		if ( ! PL_get_atom_chars( tInvokeList, & tszAtom ) )
			return false;
		tszcMethod = tszAtom;
		return true;
	}

	//  The invocation term is a list: process it
	term_t tHead = PL_new_term_ref();
	term_t tTail = PL_copy_term_ref( tInvokeList );

	bool bResult = false;
	for ( int ielem = 0; PL_get_list( tTail, tHead, tTail ); ielem++ )
	{
		switch ( ielem )
		{
			//  Zeroth element: method name
			case 0:
			{
				char * tszAtom;
				if ( ! PL_get_atom_chars( tHead, & tszAtom ) )
					return false;
				tszcMethod = tszAtom;
				bResult = true;
				break;
			}

			//  First element: atom representing invocation method
			case 1:
				dwInvoke = invokeAtom( tHead );
				break;

			default:
				break;
		}
	}
	return bResult;
}

//
//	Invoke a method on an IDispatch interface.
//
foreign_t actxInvokeObject ( 
	term_t tIDispatchFunctor, 
	term_t tInvokeList,
	term_t tArgList,
	term_t tResult )
{
	term_t tArgInvalid = 0;
	IP_IDispatch ipId;

	//  Validate the reference
	if (  ! bValidActiveXReference( tIDispatchFunctor, ipId ) )
		tArgInvalid = tIDispatchFunctor;
	//  Validate the invocation term
	else
	if ( ! (PL_is_list( tInvokeList ) || PL_is_atom( tInvokeList ) ) )
		tArgInvalid = tInvokeList;
	//  Check that the argument term is a list
	else
	if ( ! PL_is_list( tArgList ) )
		tArgInvalid = tArgList;

	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxInvokeObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	//  Parse the invocation term: it may be a single atom or a list
	TSZC tszcMethod;
	DWORD dwInvoke = 0;
	if ( ! bParseInvocationList( tInvokeList, dwInvoke, tszcMethod ) )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxInvokeObject",
						PL_CHARS, "invalid invocation list/atom" );
		return raisePrologException( tExcept );
	}

	//
	//	Set up the other information necessary for the call.
	//
	HRESULT hr = S_OK;
	ACTXDPARAMS params;
	_variant_t varResult;
	UINT uArgErr = 0;
	ACTXEXCEPINFO excepInfo;
	IP_ITypeInfo ipInfo;
	BINDPTR bindPtr;
	ZSTR zsError;
	MEMBERID mbrId;
	do
	{
		hr = hrGetInvokeInfo( ipId, 
							  tszcMethod, 
							  dwInvoke, 
							  mbrId,
							  ipInfo,
							  bindPtr );
		if ( FAILED(hr) )
		{
			zsError = "failed to get method function bind information";
			break;
		}

		switch ( dwInvoke )
		{
			case INVOKE_PROPERTYPUT:
				params.PreparePropPut(1);
				break;
		}

		hr = hrConvertArgList( ipInfo, 
								bindPtr,
								tArgList, 
								params, 
								zsError );
		if ( FAILED(hr) )
		{	
			if ( zsError.length() == 0 )
				zsError = "failed to convert argument list";

			break;
		}

		hr = ::DispInvoke( ipId,
							ipInfo,
							mbrId, 
							dwInvoke,
							& params,
							& varResult,
							& excepInfo,
							& uArgErr );

		if ( FAILED(hr) )
		{
			ZSTR zsSource("<no source>");
			ZSTR zsDesc("<no description>");

			if ( excepInfo.bstrDescription )
				zsDesc = (WSZC) excepInfo.bstrDescription;
			if ( excepInfo.bstrSource )
				zsSource = (WSZC) excepInfo.bstrSource;

			zsError.Format("IDispatch::Invoke returned HR=0x%x,SCODE=0x%x,\'%s\', (source \'%s\')", 
							hr,excepInfo.scode,
							zsDesc.Tszc(),
							zsSource.Tszc() );
			break;
		}

		if ( ! bUnifyVariantWithTerm( varResult, tResult ) )
		{
			zsError = "failed to unify result with result term";
			hr = E_FAIL;
			break;
		}

		hr = S_OK;
	} 
	while ( false );

	if ( FAILED(hr) )
	{
		if ( g_bDispatchErrorsAsExceptions )
		{
			term_t tExcept = PL_new_term_ref();
			PL_unify_term( tExcept,
							PL_FUNCTOR, fnctrActxError, 
							PL_CHARS, "actxInvokeObject",
							PL_CHARS, zsError.Tszc(),
							PL_TERM, tIDispatchFunctor );
			return raisePrologException( tExcept );
		}
		else
		{
			HRFAIL(hr);
		}
	}

	PL_succeed;
}

//
//	Perform generic automation object enumeration.
//
//      This is done through backtracking across each enumerated value.
//      A special ACTXTOKEN is captured to represent any IDispatch pointers
//      which may result from the enumeration; this interface pointer will
//      be overwritten during REDO and discarded finally during CUT.
//
foreign_t actxEnumObject (
	term_t tIDispatchFunctor, 
	term_t tResult,
	foreign_t fHand )
{
	int iCallType = PL_foreign_control(fHand);
	term_t tArgInvalid = 0;
	IP_IDispatch ipId;

	if (   iCallType == PL_FIRST_CALL 
		&& (! bValidActiveXReference( tIDispatchFunctor, ipId )) )
		tArgInvalid = tIDispatchFunctor;

	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxEnumObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	BKTRKENUM * pbktrk = NULL;
	HRESULT hr = S_OK;

	switch ( iCallType )
	{       
		case PL_FIRST_CALL:
		{
			//  Create the backtracking data storage object
			pbktrk = new BKTRKENUM();
			pbktrk->_atkn = actxJournal.nextToken();
			//  Prepare for the IDispatch::Invoke
			ACTXDPARAMS params;
			_variant_t varResult;
			UINT uArgErr = 0;
			ACTXEXCEPINFO excepInfo;
			params.resize(0);

			IP_ITypeInfo ipInfo;
			hr = ipId->GetTypeInfo( 0, g_lcid, & ipInfo );
			if ( FAILED(hr) )
				HRFAIL(hr);

			hr = ::DispInvoke( ipId,
								ipInfo,
								DISPID_NEWENUM,
								DISPATCH_METHOD | DISPATCH_PROPERTYGET,
								& params,
								& varResult,
								& excepInfo,
								& uArgErr );
			if ( FAILED(hr) )
			{
				HRFAIL(hr);
			}
			else
			{
				pbktrk->_ipEnum = varResult;
			}
			break;
		}

		case PL_REDO:
		{
			pbktrk = (BKTRKENUM*) PL_foreign_context_address(fHand);
			break;
		}

		case PL_CUTTED:
		{
			pbktrk = (BKTRKENUM*) PL_foreign_context_address(fHand);
			delete pbktrk;
			PL_succeed;
			break;
		}

		default:
			assert("unknown non-determ call type in actxEnumObject");
			HRFAIL(E_PROLOG_FAIL);
			break;
	}
	if ( ! pbktrk )
		HRFAIL(E_PROLOG_FAIL);

	//  Access the next element of the enumeration
	IEnumVARIANTPtr & ipEnum = pbktrk->_ipEnum;
	if ( ! ipEnum )
	{
		delete pbktrk;
		HRFAIL(E_PROLOG_FAIL);
	}

	_variant_t varNext;
	ULONG ulReturned = 0;
	hr = ipEnum->Next( 1, & varNext, & ulReturned );
	if ( FAILED(hr) || varNext.vt == VT_EMPTY )
	{
		delete pbktrk;
		HRFAIL(hr);
	}

	if ( ! bUnifyVariantWithTerm( varNext, tResult, & pbktrk->_atkn ) )
	{
		delete pbktrk;
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxEnumObject",
						PL_CHARS, "cannot convert returned enum VARIANT to Prolog term",
						PL_TERM, tIDispatchFunctor );
		return raisePrologException( tExcept );
	}

	PL_retry_address(pbktrk);
}

//      
//      Get the "Count" property value from a standard collection.
//
HRESULT hrCollectionCount ( 
	IP_IDispatch & ipId,
	int & iCount )
{
	MEMBERID mbrId;

	IP_ITypeInfo ipInfo;
	BINDPTR bindPtr;

	HRESULT hr = hrGetInvokeInfo( ipId, 
								  "Count", 
								  INVOKE_PROPERTYGET, 
								  mbrId,
								  ipInfo,
								  bindPtr );
	if ( FAILED(hr) )
		return hr; 

	iCount = 0;
	ACTXDPARAMS params;
	_variant_t varResult;
	UINT uArgErr = 0;
	ACTXEXCEPINFO excepInfo;
	params.resize(0);

	do
	{

		hr = ::DispInvoke( ipId,
							ipInfo,
							mbrId, 
							INVOKE_PROPERTYGET,
							& params,
							& varResult,
							& excepInfo,
							& uArgErr );

		if ( ! bConvertVariantToInt ( varResult, iCount ) )
		{
			hr = E_FAIL;
			break;
		}
	}
	while (false);

	return hr;
}

//
//	Perform generic automation object enumeration.  Result
//      is a list.
//
foreign_t actxCollectionList ( 
	term_t tIDispatchFunctor, 
	term_t tResult )
{
	term_t tArgInvalid = 0;
	IP_IDispatch ipId;

	if ( ! bValidActiveXReference( tIDispatchFunctor, ipId ) )
		tArgInvalid = tIDispatchFunctor;
	else
	if ( ! PL_is_variable( tResult ) )
		tArgInvalid = tResult;

	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxCollectionList",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	//  Get the item count from the collection
	int cItems;
	HRESULT hr = hrCollectionCount( ipId, cItems );

	if ( FAILED(hr) )
	{
		HRFAIL(hr);
	}

	//  Build the list tail-to-head
	term_t tList = PL_new_term_ref();
	PL_put_nil( tList );

	ACTXDPARAMS params;
	UINT uArgErr = 0;
	ACTXEXCEPINFO excepInfo;
	IP_ITypeInfo ipInfo;

	hr = ipId->GetTypeInfo( 0, 0, & ipInfo );
	if ( FAILED(hr) )
	{
		HRFAIL(hr);
	}

	for ( int iItem = cItems; --iItem >= 0; )
	{
		_variant_t varResult;

		params.resize(1);
		params.VVars()[0] = (long) iItem;

		hr = ::DispInvoke(  ipId, 
							ipInfo,
							DISPID_VALUE,
							INVOKE_PROPERTYGET,
							& params,
							& varResult,
							& excepInfo,
							& uArgErr );
		if ( FAILED(hr) )
			break;

		term_t tHead = PL_new_term_ref();
		if ( ! bUnifyVariantWithTerm( varResult, tHead ) )
		{
			term_t tExcept = PL_new_term_ref();
			PL_unify_term( tExcept,
							PL_FUNCTOR, fnctrActxError, 
							PL_CHARS, "actxCollectionList",
							PL_CHARS, "cannot convert returned enum VARIANT to Prolog term",
							PL_TERM, tIDispatchFunctor );
			return raisePrologException( tExcept );
		}
		PL_cons_list( tList, tHead, tList );
	}

	if ( FAILED(hr) )
	{
		HRFAIL(hr);
	}
	if (  ! PL_unify( tList, tResult ) )
		HRFAIL(E_PROLOG_FAIL);

	PL_succeed;
}

foreign_t actxCloneObject ( 
	term_t tIDispatchFunctor, 
	term_t tResult )
{
	term_t tArgInvalid = 0;
	IP_IDispatch ipId;

	if (  ! bValidActiveXReference( tIDispatchFunctor, ipId ) )
		tArgInvalid = tIDispatchFunctor;
	else
	if ( ! PL_is_variable( tResult ) )
		tArgInvalid = tResult;

	if ( tArgInvalid )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxCloneObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tArgInvalid );
		return raisePrologException( tExcept );
	}

	ACTXTOKEN atkn = actxJournal.registerObject( ipId );
	PL_unify_term( tResult,
					PL_FUNCTOR, fnctrActxObject, 
					PL_INTEGER, atkn );
	PL_succeed;
}


foreign_t actxReleaseAll ()
{
	actxJournal.clear();
	PL_succeed;
}

foreign_t actxBreakCheck()
{
#ifdef _DEBUG
	_CrtDbgBreak();
#endif
	PL_succeed;
}


//
//      Create a non-deterministic predicate that succeeds exactly once.
//      	This creates a "context stack point" and allows us to remove
//      	all registered ActiveX objects created after this point.
//
foreign_t actxContext ( foreign_t fHand )
{
	int iCallType = PL_foreign_control(fHand);
	BKTRKCONTEXT * pbkcon = NULL;

	switch ( iCallType )
	{       
		//  Allocate the context block and succeed on the first call
		case PL_FIRST_CALL:
		{
			pbkcon = new BKTRKCONTEXT;
			PL_retry_address(pbkcon);
			break;
		}

		//  Recover the context block and fail others
		case PL_REDO:
		case PL_CUTTED:
		default:
		{
			pbkcon = (BKTRKCONTEXT *) PL_foreign_context_address(fHand);
			assert( pbkcon );
			delete pbkcon;
			PL_fail;
			break;
		}
	}
}

//
//      Set or determine the context of an activeXObject.
//
//      	If 'tBool' is instantiated to the atom 'true', set
//      		the context of the object to the global context;
//      	If 'tBool' is instantiated to the atom 'false', set
//      		the context to the most recent local context.
//      	Otherwise, attempt to unify 'tBool' with the atom
//      		representing the object's current context.
//
foreign_t actxContextGlobal ( 
	term_t tIDispatchFunctor,
	term_t tBool )
{
	IP_IDispatch ipId;
	ACTXTOKEN tknObj;

	if ( ! bValidActiveXReference( tIDispatchFunctor, ipId, & tknObj ) )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxContextObject",
						PL_CHARS, "invalid argument",
						PL_TERM, tIDispatchFunctor );
		return raisePrologException( tExcept );
	}

	ACTXITEM * paitem = actxJournal.PActxItem( tknObj );
	int iContext = paitem->IContext();
	atom_t atm = iContext >= 0 
			 ? atomTrue 
			 : atomFalse;

	//  If argument is instantiated to an atom, 
	//      	set context to global or local accordingly.
	//
	//      	'true'  ==> unbind (treat as global)
	//      	'false' ==> bind to current local context
	//
	if ( PL_is_atom( tBool ) )
	{
		atom_t atomBool;
		PL_get_atom( tBool, & atomBool );

		if ( atomBool != atomTrue && atomBool != atomFalse )
			PL_fail;

		actxJournal.bSetContext( tknObj, atomBool == atomTrue );
		PL_succeed;
	}
	else
	{
		return PL_unify_atom( tBool, atm );
	}
}

//
//	Set or return the "treat errors as exceptions" flag
//
foreign_t actxErrorsAsExceptions ( term_t tBool )
{
	atom_t atm = g_bDispatchErrorsAsExceptions
			 ? atomTrue 
			 : atomFalse;

	if ( PL_is_atom( tBool ) )
	{
		atom_t atomBool;
		if ( ! PL_get_atom( tBool, & atomBool ) )
			PL_fail;

		if ( atomBool != atomTrue && atomBool != atomFalse )
			PL_fail;

		g_bDispatchErrorsAsExceptions = atomBool == atomTrue;
		PL_succeed;
	}
	else
	{
		return PL_unify_atom( tBool, atm );
	}
}

foreign_t actxListToDate ( term_t tList, term_t tDate )
{
	bool bVarList = PL_is_variable( tList );
	bool bVarDate = PL_is_variable( tDate );
	DATE date = 0;

	//	Raise an exception if both arguments unbound
	if ( bVarList && bVarDate )
	{
		term_t tExcept = PL_new_term_ref();
		PL_unify_term( tExcept,
						PL_FUNCTOR, fnctrActxError, 
						PL_CHARS, "actxListToDate",
						PL_CHARS, "both arguments unbound",
						PL_TERM, tList );
		return raisePrologException( tExcept );
	}
	else
	//  If list is bound, convert it to a date
	if ( ! bVarList )
	{	
		//  Convert the list to a DATE
		if ( ! bConvertListToDate( tList, date ) )
		{
			term_t tExcept = PL_new_term_ref();
			PL_unify_term( tExcept,
							PL_FUNCTOR, fnctrActxError, 
							PL_CHARS, "actxListToDate",
							PL_CHARS, "list-to-date conversion failed",
							PL_TERM, tDate );
			return raisePrologException( tExcept );
		}

		double ddate = 0;
		if ( PL_is_float( tDate ) || PL_is_variable( tDate ) )
		{
			//  Attempt to unify with the given float
			if ( ! PL_unify_float( tDate, date ) )
				PL_fail;
		}
		else
		if ( PL_is_integer( tDate ) )
		{
			if ( ! bIntegralFloat( date ) )
				PL_fail;
			int i = date;
			//  Attempt to unify with the given integer
			if ( ! PL_unify_integer( tDate, i ) )
				PL_fail;
		}
		else
			PL_fail;

	}
	else
	{
		//  List is unbound, so get date...
		if ( ! bGetDblFromTerm( tDate, date ) )
			PL_fail;
		//		and unify with list
		if ( ! bUnifyDateWithList( date, tList ) )
			PL_fail;
	}
	PL_succeed;
}



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
This function is a handle  called   from  abort/1.   The function should
perform cleanup as Prolog is going to   perform a long_jmp() back to the
toplevel.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void
actx_abort(void)
{ 
	actxJournal.clear();
}



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
(un)install functions.  Predicates registered with PL_register_foreign()
do not need to be uninstalled, as the Prolog toplevel driver
unload_foreign_library/[1,2] will to this automatically for you.

As only hooks need to be uninstalled,  you won't need this function very
often.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

install_t
install()
{ 
	//  Check if we've already done this
	if ( g_bInstalled )
		return;

	g_bInstalled = true;

	//  Register our constant strings
	init_constants();
	//  Inform us when an abort is taken so that all objects may be cleared
	PL_abort_hook(actx_abort);

	//  Register our functions
	PL_register_foreign("actxCreateObject", 	2,      actxCreateObject,		0);
	PL_register_foreign("actxReleaseObject",	1,      actxReleaseObject,      0);
	PL_register_foreign("actxInvokeObject", 	4,      actxInvokeObject,       0);
	PL_register_foreign("actxQueryInterface",   3,      actxQueryInterface,     0);
	PL_register_foreign("actxEnumObject",		2,      actxEnumObject,		 	PL_FA_NONDETERMINISTIC);
	PL_register_foreign("actxCloneObject",  	2,      actxCloneObject,		0);
	PL_register_foreign("actxCollectionList",   2,      actxCollectionList,		0);
	PL_register_foreign("actxReleaseAll",   	0,      actxReleaseAll, 		0);
	PL_register_foreign("actxBreak",			0,      actxBreakCheck, 		0);
	PL_register_foreign("actxContext",      	0,      actxContext,    		PL_FA_NONDETERMINISTIC);
	PL_register_foreign("actxContextGlobal",	2,      actxContextGlobal,      0);
	PL_register_foreign("actxListToDate",		2,      actxListToDate,			0);
	PL_register_foreign("actxErrorsAsExceptions",1,     actxErrorsAsExceptions,	0);
}


install_t
uninstall()
{ 
	PL_abort_unhook(actx_abort);
}

SWIACTX_RESIDENT ACTXJOURNAL * PSwiActxGetJournal ()
{
	install();

	return & actxJournal;
}


