/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	swixcvt.cpp
//
#pragma warning ( disable : 4786 )      	//  overlength debug info 

#include "swiactx.h"
#include "olehelp.h"

//
//	Class Id and guid conversion functions
//
bool bStringFromGuid ( const GUID & guid, ZSTR & zsGuid )
{
	LPOLESTR pszGUID = NULL;
	HRESULT hr = ::StringFromCLSID( guid, &pszGUID );
	if ( FAILED(hr) )
		return false;
	zsGuid = pszGUID;
	::CoTaskMemFree(pszGUID);
	return true;
}

bool bGuidFromString ( TSZC tszcGuid, GUID & guid )
{
	ZSTRW zswGuid( tszcGuid );
	HRESULT hr = ::CLSIDFromString( const_cast<WCHAR*>(zswGuid.Tszc()), & guid );
	return SUCCEEDED(hr);
}

bool bGetClsId ( TSZC tszcProgId, CLSID & clsid )
{
	ZSTRW zswProgId( tszcProgId );
	LONG lResult = CLSIDFromProgID( const_cast<WCHAR*>(zswProgId.Tszc()), & clsid );
	return lResult == 0;
}

bool bGetClsIdString ( TSZC tszcProgId, ZSTR & zsClsId )
{
	CLSID clsid;
	if ( ! bGetClsId( tszcProgId, clsid ) )
		return false;
	return bStringFromGuid( clsid, zsClsId );
}

//  Extract a double from an integer or float term
bool bGetDblFromTerm ( term_t tTerm, double & dbl )
{
	switch ( PL_term_type( tTerm ) )
	{
		case PL_INTEGER:
		{
			int i;
			if (! PL_get_integer( tTerm, & i ) )
				return false;
			dbl = i;
			break;
		}
		case PL_FLOAT:
		{
			if (! PL_get_float( tTerm, & dbl ) )
				return false;
			break;
		}
		default:
			return false;
	}
	return true;
}

//  Extract an integer from an integer or float term; float must
//		be integral.
bool bGetIntFromTerm ( term_t tTerm, int & i )
{
	switch ( PL_term_type( tTerm ) )
	{
		case PL_INTEGER:
		{
			if (! PL_get_integer( tTerm, & i ) )
				return false;
			break;
		}
		case PL_FLOAT:
		{
			double dbl;
			if (! PL_get_float( tTerm, & dbl ) )
				return false;
			if ( ! bIntegralFloat( dbl ) )
				return false;
			i = dbl;
			break;
		}
		default:
			return false;
	}
	return true;
}

//	Determine if a double value is integral.
bool bIntegralFloat ( const double & dbl )
{
	int i = dbl;
	double dbl2 = i;
	return dbl == dbl2;
}

//  Convert a VARIANT to an integer; supports VT_{U}I{2,4,INT}, among others
bool bConvertVariantToInt ( _variant_t & vnt, int & iResult )
{	
	iResult = 0;
	bool bResult = true;
	switch ( vnt.vt )
	{		
		case VT_UI2:
		{
			iResult = vnt.uiVal;
			break;
		}

	    case VT_I2:
		{
			iResult = vnt.iVal;
			break;
		}

		case VT_UI4:
		case VT_UINT:
		{
			iResult = vnt.ulVal;
			break;
		}

		case VT_INT:
		case VT_I4:
		{
			iResult = vnt.lVal;
			break;
		}

		case VT_BOOL:
		{
			iResult = vnt.iVal == VARIANT_TRUE;
			break;
		}

		case VT_EMPTY:
			iResult = 0;
			break;

		default:
			bResult = false;
			break;
			
	}
	return bResult;
}

//	Scan a list in order to see if its elements are compatible; returns
//		the count of items in the list and their VARTYPE.
//
//		If input VARTYPE is VT_EMPTY, it will be set to the common element type,
//			and all elements must be of that type.
//		If input VARTYPE is VT_VARIANT, any element type is allowed.
//		If input VARTYPE is other, all elements must be of that type.
//
bool bPrepareList ( 
	term_t tList, 
	VARTYPE & vt, 
	int & citems )
{
	citems = 0;
	term_t tHead = PL_new_term_ref();
	term_t tTail = PL_copy_term_ref(tList);

	while ( PL_get_list( tTail, tHead, tTail ) )
	{
		VARTYPE vtItem = VT_NULL;
		//  Get the type of this term
		int iarg = PL_term_type( tHead );

		if ( PL_is_string( tHead ) )
			vtItem = VT_BSTR;
		else
		if ( PL_is_atom( tHead ) )
			vtItem = VT_BSTR;
		else
		if ( PL_is_integer( tHead ) )
		{
			// If the type is already an integer type, don't overstore it.
			switch ( vt )
			{
				case VT_I1:
				case VT_UI1:
				case VT_I2:
				case VT_UI2:
				case VT_I4:
				case VT_UI4:
				case VT_INT:
				case VT_UINT:
					vtItem = vt;
					break;
				default:
					vtItem = VT_I4;
					break;
			}
		}
		else
		if ( PL_is_float( tHead ) )
			vtItem = VT_R8;
		else
		//BUGBUG! Force anything else to be a string
		if ( ! PL_is_variable( tHead ) )
			vtItem = VT_BSTR;	

		//  Can we handle this type? 
		//		(don't allow unbound variables in lists)
		if ( vtItem == VT_NULL )
			return false;

		//  If list is still of unassigned type, 
		if ( vt == VT_EMPTY )
			vt = vtItem;		// use this type
		else	
		// else, verify that it's of the same type or the
		//	base type is known to be VT_VARIANT
		if ( vt != vtItem && vt != VT_VARIANT )
			return false;

		citems++;
	}
	return true;
}

//
//	Convert a list of Prolog terms to a SAFEARRAY nestled in a VARIANT.
//	We do not allow unbound terms in a list.  All elements of a list must
//	be of the same type unless the VARTYPE is VT_VARIANT.
//
bool bConvertListToVariantSafearray ( 
	term_t tList,						//  The list head
	_variant_t & vnt,					//	The resulting variant containing the SAFEARRAY
	VARTYPE vt /* = VT_EMPTY */ )		//	The desired VARTYPE; VT_EMPTY implies "guess"
{
	int citems = 0;

	//  Check that all the elements of the list are of the same type
	if ( ! bPrepareList( tList, vt, citems ) )
		return false;

	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].lLbound = 0;
	rgsabound[0].cElements = citems;

	vnt.vt = VT_ARRAY | vt;
	HRESULT hr = S_OK;
	SAFEARRAY * psfa = NULL;

	double dbl;
	int i;
	ZSTRW zsw;
	void * pv;
	do
	{
		SAFEARRAY * psfa = ::SafeArrayCreate( vt, 1, rgsabound );
		if ( ! psfa )
		{
			hr = E_FAIL;
			break;
		}
		vnt.parray = psfa;
		hr = ::SafeArrayAllocData( psfa );
		if ( FAILED(hr) )
			break;
		
		term_t tHead = PL_new_term_ref();
		term_t tTail = PL_copy_term_ref(tList);

		for ( long iIndex = 0; 
			  SUCCEEDED(hr) && PL_get_list( tTail, tHead, tTail ); 
			  iIndex++ )
		{			
			switch ( vt )
			{
				case VT_BSTR:
				{
					char * psz;
					if ( ! PL_get_chars( tHead, & psz, CVT_ALL ) )
					{
						hr = E_FAIL;
						break;
					}
					zsw = psz;
					_bstr_t bstr = zsw;
					pv = bstr.copy();
					break;
				}

				case VT_I1:
				case VT_UI1:
				case VT_I2:
				case VT_UI2:
				case VT_I4:
				case VT_UI4:
				case VT_INT:
				case VT_UINT:
				{
					PL_get_integer( tHead, & i );
					pv = & i;
					break;
				}
				case VT_R8:
				{
					PL_get_float( tHead, & dbl );
					pv = & dbl;
					break;
				}

				case VT_VARIANT:
				{
					//  This should never change
					assert( sizeof(VARIANT) == sizeof(_variant_t) );
					//  Construct a new _variant_t in the SAFEARRAY data
				    _variant_t * pvar = (_variant_t *) pv;
					new (pvar) _variant_t;
					//  Convert the next term to a variant
					if ( ! bConvertTermToVariant( VT_VARIANT, tHead, *pvar ) )
						hr = E_FAIL;
					break;
				}

				default:
				{
					assert(false);
					hr = E_FAIL;
					break;
				}
			}
			if ( SUCCEEDED(hr) )
				hr = ::SafeArrayPutElement( psfa, & iIndex, pv );
		}
	} 
	while (false);

	if ( FAILED(hr) )
	{
		if ( psfa )
			::SafeArrayDestroy( psfa );
		vnt.parray = NULL;
		vnt.Clear();
	}
	return SUCCEEDED(hr);
}

//  Convert an activeXVariant functor to a VARIANT.
bool bConvertVariantFunctorToVariant ( 
	term_t tData,						//  The data term
	_variant_t & vnt,					//	The resulting variant containing the data
	VARTYPE vt /* = VT_EMPTY */ )		//	The desired VARTYPE; VT_EMPTY implies "guess"
{
	bool bResult = false;

	switch ( vt )
	{
		case VT_DATE:
		{
			vnt.vt = VT_DATE;
			double d;
			if ( ! PL_get_float( tData, & d ) )
				break;
			vnt.date = d;
			bResult = true;
			break;
		}

		default:
			break;
	}
	return bResult;
}

//
//	Create a list containing the integer terms of the given date,
//		then attempt to unify the result with the given term.
//
bool bUnifyDateWithList ( const DATE & date, term_t tListResult )
{
	tm tmDate;
	term_t tList = PL_new_term_ref();
	PL_put_nil( tList );
	term_t tInt;
	bool bResult = false;
	do
	{
		if ( ! TmFromOleDate( date, tmDate ) )
			break;
		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_sec ) )
			break;
		PL_cons_list( tList, tInt, tList );

		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_min ) )
			break;
		PL_cons_list( tList, tInt, tList );

		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_hour ) )
			break;
		PL_cons_list( tList, tInt, tList );

		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_mday ) )
			break;
		PL_cons_list( tList, tInt, tList );

		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_mon ) )
			break;
		PL_cons_list( tList, tInt, tList );

		tInt = PL_new_term_ref();
		if ( ! PL_unify_integer( tInt, tmDate.tm_year ) )
			break;
		PL_cons_list( tList, tInt, tList );

		if ( ! PL_unify( tList, tListResult ) )
			break;
		bResult = true;
	}
	while ( false );
	return bResult;
}

//  Convert a list of integers to a DATE.
bool bConvertListToDate ( term_t tList, DATE & date )
{
	if ( ! PL_is_list( tList ) )
		return false;

	term_t tHead = PL_new_term_ref();
	term_t tTail = PL_copy_term_ref( tList );
	WORD w[6];
	int i;

	for ( int iw = 0; iw < 6; iw++ )
	{
		if ( ! PL_get_list( tTail, tHead, tTail ) )
			break;
		if ( ! PL_get_integer( tHead, & i ) )
			break;
		w[iw] = i;
	}
	if ( iw != 6 )
		return false;

	return OleDateFromTm( w[0], w[1], w[2], w[3], w[4], w[5], date );
}


//
//	Determine if the given term is a valid functor for our ActiveX layer.
//  	In Prolog, they look like "activeXObject(n)", where n is the index in
//     	the "actxJournal" map.
//
bool bValidActiveXReference ( 
	term_t tIDispatchFunctor, 
	IP_IDispatch & ipId, 
	ACTXTOKEN * piObjId )
{
	functor_t fnctr;
	term_t tObjectId = PL_new_term_ref();

	bool bOK = false;
	do 
	{
		if (! PL_is_functor( tIDispatchFunctor, fnctrActxObject )) 
			break;
		if (! PL_get_functor( tIDispatchFunctor, & fnctr ))
			break;
		if ( fnctr != fnctrActxObject ) 
			break;
		 if (! PL_get_arg( 1, tIDispatchFunctor, tObjectId ))
			break;
		 if ( ! PL_is_integer( tObjectId ) )
			break;
		 int iObj;
		 if ( ! PL_get_integer( tObjectId, & iObj ) )
			break;
		 if ( piObjId )
			*piObjId = iObj;
		 if (! actxJournal.bFind( iObj, ipId ) )
			break;
		bOK = true;
	}
	while ( false );
	return bOK;
}

//
//  Validate that a term is a SAFEARRAY wrapper funtor, 
//		which looks like:
//
//			activeXSafearray(VariantType,ListOfElements).
//
bool bValidActiveXSafearray ( 
	term_t tSafearray, 
	VARTYPE & vt,
	term_t & tList )
{
	functor_t fnctr;
	term_t tVartype = PL_new_term_ref();
	
	int ivt;

	bool bOK = false;
	do 
	{
		//  Test for expected functor
		if (! PL_is_functor( tSafearray, fnctrActxSafearray )) 
			break;
		if (! PL_get_functor( tSafearray, & fnctr ))
			break;
		if (fnctr != fnctrActxSafearray) 
			break;

		//  Test for and extract arg #1, the VARTYPE integer
		if (! PL_get_arg( 1, tSafearray, tVartype ))
			break;
		if ( ! PL_is_integer( tVartype ) )
			break;
		if ( ! PL_get_integer( tVartype, & ivt ) )
			break;

		//  Test for and extract arg #2, the list of arguments
		tList = PL_new_term_ref();
		if (! PL_get_arg( 2, tSafearray, tList ))
			break;
		if ( ! PL_is_list( tList ) )
			break;

		//  Success: update the VARTYPE
		vt = ivt;
		bOK = true;
	}
	while ( false );
	return bOK;
}

//
//  Validate that a term is a VARIANT wrapper funtor, 
//		which looks like:
//
//			activeXVariant(VariantType,DataTerm).
//
bool bValidActiveXVariant ( 
	term_t tVariant, 
	VARTYPE & vt,
	term_t & tData)
{
	functor_t fnctr;
	term_t tVartype = PL_new_term_ref();
	
	int ivt;

	bool bOK = false;
	do 
	{
		//  Test for expected functor
		if (! PL_is_functor( tVariant, fnctrActxVariant )) 
			break;
		if (! PL_get_functor( tVariant, & fnctr ))
			break;
		if ( fnctr != fnctrActxVariant ) 
			break;

		//  Test for and extract arg #1, the VARTYPE integer
		if (! PL_get_arg( 1, tVariant, tVartype ))
			break;
		if ( ! PL_is_integer( tVartype ) )
			break;
		if ( ! PL_get_integer( tVartype, & ivt ) )
			break;

		//  Test for and extract arg #2, the variant data
		tData = PL_new_term_ref();
		if (! PL_get_arg( 2, tVariant, tData ))
			break;

		//  Success: update the VARTYPE
		vt = ivt;
		bOK = true;
	}
	while ( false );
	return bOK;
}


//
//	Unify a single element of a SAFEARRAY with a term.
//		Element is represented by VARTYPE and pointer to data.
//
bool bConvertSafeArrayElementToTerm (
	VARTYPE vt,
	void * pv,
	term_t tElem )
{
	bool bResult = true;
	IP_IDispatch ipId;

	switch ( vt )
	{
		//  One byte integers
		case VT_I1:
		{
			int i = *((char*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		case VT_UI1:
		{
			int i = *((unsigned char*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		//  Two byte integers
		case VT_UI2:
		{
			int i = *((unsigned short*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		case VT_I2:
		{
			int i = *((short*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		//  Four byte integers
		case VT_INT:
		case VT_UI4:
		case VT_UINT:
		case VT_I4:
		{
			int i = *((LONG*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		case VT_R4: 
		{
			double d = *((float*)pv);
			PL_put_float( tElem, d );
			break;
		}

		case VT_R8:
		{
			double d = *((double*)pv);
			PL_put_float( tElem, d );
			break;
		}

		case VT_BSTR:
		{
			ZSTR zs( *((BSTR*) pv) );
			PL_put_string_chars( tElem, zs );
			break;
		}

		case VT_BOOL:
		{
			atom_t atm = *((SHORT*)pv) == VARIANT_TRUE
					 ? atomTrue
					 : atomFalse;
			PL_put_atom( tElem, atm );
			break;
		}

		case VT_DISPATCH:
		{
			ipId = *((IDispatch **) pv);
			if ( ! ipId )
				return false;
			ACTXTOKEN atkn = actxJournal.registerObject( ipId );
			bResult = PL_unify_term( tElem,
									PL_FUNCTOR, fnctrActxObject, 
									PL_INTEGER, atkn );     
			break;
		}

		default:
			return false;
	}

	return bResult;
}

//
//	Convert a dimension of a SAFEARRAY into a Prolog list.
//	If this is not the lowest dimension, create a list of lists.
//
bool bConvertSafeArrayDimensionToList ( 
	VARTYPE vt,
	SAFEARRAY * psfa,
	VLONG & vidim,
	int iDim,
	term_t tList )
{
	LONG iLbound = 0, 
		 iUbound = 0,
		 iItem = 0;

	HRESULT hr = S_OK;

	int iDimThis = vidim.size() - iDim;
	vidim[ iDimThis ] = 0;

	int cbElem = ::SafeArrayGetElemsize( psfa );    
	void * pv = _alloca( cbElem ); 
	do 
	{

		hr = ::SafeArrayGetLBound( psfa, iDim, & iLbound );
		if ( FAILED(hr) )
			break;
		hr = ::SafeArrayGetUBound( psfa, iDim, & iUbound );
		if ( FAILED(hr) )
			break;

		for ( iItem = iUbound; iItem >= iLbound; iItem-- )
		{
			term_t tElem = PL_new_term_ref();
			vidim[iDimThis] = iItem;

			//  Is this the lowest (fastest varying) dimension)?
			if ( iDimThis == 0 )
			{
				//  Yes- start extracting and converting elements
				hr = ::SafeArrayGetElement( psfa, & vidim[0], pv );
				if ( FAILED(hr) )
					break;

				if ( ! bConvertSafeArrayElementToTerm( vt, pv, tElem ) )
					break;
			}
			else
			{
				//  No- this is an intermediate dimension
				if ( ! bConvertSafeArrayDimensionToList( vt, psfa, vidim, iDim+1, tElem ) )
					break;

			}
			PL_cons_list( tList, tElem, tList );
		}

		// Check that we processed all the items
		if ( SUCCEEDED(hr) && iItem >= iLbound )
			hr = E_FAIL;
	} 
	while (false);

	return SUCCEEDED(hr);
}



//
//	Unify an array variant with a list.   
//
bool bUnifyVariantArrayWithTerm ( 
	_variant_t & vnt,   	//  VARIANT to convert
	term_t & tTerm, 		//  Term to unify with 
	ACTXTOKEN * patkn )      //  Optional token for interface registration
{
	assert( vnt.vt & VT_ARRAY );

	//  Start the list 
	term_t tList = PL_new_term_ref();
	PL_put_nil( tList );

	SAFEARRAY * psfa = vnt.parray;
	assert( psfa );

	VARTYPE vt = vnt.vt - VT_ARRAY;

	VLONG vidim;
	int cdim = ::SafeArrayGetDim( psfa );
	vidim.resize( cdim );

	bool bResult = bConvertSafeArrayDimensionToList( vt, psfa, vidim, 1, tList );
	return bResult && PL_unify( tList, tTerm );
}

//
//	Create the special SAFEARRAY descriptor functor along with its list
//		and unify it with the given term.
//
bool bUnifyVariantArrayWithSafearrayFunctor ( 
	_variant_t & vnt, 
	term_t & tTerm,
	ACTXTOKEN * patkn )      //  Optional token for interface registration
{
	int ivt = vnt.vt;
	if ( (ivt & VT_ARRAY) == 0 )
		return false;		//  Not an array
	ivt -= VT_ARRAY;

	term_t tList = PL_new_term_ref();
	if ( ! bUnifyVariantArrayWithTerm( vnt, tList, patkn ) )
		return false;		//  Unable to create list

	return PL_unify_term( tTerm,
							PL_FUNCTOR, fnctrActxSafearray, 
							PL_INTEGER, ivt,
							PL_TERM, tList );

}

//
//	Convert a returned VARIANT to a Prolog term.  
//
bool bUnifyVariantWithTerm ( 
	_variant_t & vnt,       	//  VARIANT to convert
	term_t & tTerm, 			//  Term to unify with 
	ACTXTOKEN * patkn )			//  Optional token for interface registration
{       
	switch ( vnt.vt )
	{
		//  Unify "empty" with "[]" (the empty list).
		case VT_NULL:
		case VT_EMPTY:
		{
			return PL_unify_nil( tTerm );
			break;
		}

		//  Two-byte integers
		case VT_UI2:
	    case VT_I2:
		{
			int i = vnt.vt == VT_UI2
				  ? vnt.uiVal 
				  : vnt.iVal;
			return PL_unify_integer( tTerm, i );
			break;
		}

		case VT_UI4:
		case VT_UINT:
		{
			int i = vnt.ulVal;
			return PL_unify_integer( tTerm, i );
			break;
		}

		//BUGBUG:  The only user-defined type we allow is ENUMs
		case VT_USERDEFINED:
		case VT_INT:
		case VT_I4:
		{
			int i = vnt.lVal;
			return PL_unify_integer( tTerm, i );
			break;
		}

		case VT_R4:
		{
			double d = vnt.fltVal;
			return PL_unify_float( tTerm, d );
			break;
		}

		case VT_R8:
		{
			double d = vnt.dblVal;
			return PL_unify_float( tTerm, d );
			break;
		}

		case VT_BSTR:
		{
			ZSTR zs( vnt.bstrVal );
			return PL_unify_string_chars( tTerm, zs );
			break;
		}

		case VT_BOOL:
		{
			atom_t atm = vnt.iVal == VARIANT_TRUE 
					   ? atomTrue 
					   : atomFalse;

			return PL_unify_atom( tTerm, atm );
			break;
		}

		case VT_DISPATCH:
		{
			IP_IDispatch ipId( vnt.pdispVal );
			ACTXTOKEN atkn = actxJournal.registerObject( ipId, patkn );
			return PL_unify_term( tTerm,
									PL_FUNCTOR, fnctrActxObject, 
									PL_INTEGER, atkn );
			break;
		}

		case VT_DATE:
		{
			return bUnifyVariantWithVariantFunctor( vnt, tTerm, patkn );

		/*  BUGBUG: use or lose: old code to stringify dates

			tm tmOle;
			if ( ! TmFromOleDate( vnt.date, tmOle ) )
				return false;
			tmOle.tm_year -= 1900;
			ZSTR zsTime( asctime(& tmOle) );
			return PL_unify_string_chars( tTerm, zsTime );
		*/
			break;
		}

		default:
		{
			if ( vnt.vt & VT_ARRAY )
				// BUGBUG:  Always unify returned arrays with the special SAFEARRAY
				//		wrapper functor.
				//	return bUnifyVariantArrayWithTerm( vnt, tTerm, patkn );
				return bUnifyVariantArrayWithSafearrayFunctor( vnt, tTerm, patkn );
			return false;
		}
	}
	return true;
}

//
//	Create the special VARIANT functor used for unusual data types
//		and unify it with the given term.
//
bool bUnifyVariantWithVariantFunctor (
	_variant_t & vnt, 
	term_t & tTerm,
	ACTXTOKEN * patkn )      //  Optional token for interface registration
{
	int ivt = vnt.vt;

	term_t tVariant = PL_new_term_ref();
	switch ( ivt )
	{
		case VT_DATE:
		{
			double d = vnt.date;
			if ( ! PL_unify_float( tVariant, d ) )
				return false;	// should never happen
			break;
		}

		default:
			return false;
	}

	return PL_unify_term( tTerm,
							PL_FUNCTOR, fnctrActxVariant, 
							PL_INTEGER, ivt,
							PL_TERM, tVariant );

}


//
//	Convert the given Prolog term to the desired VARIANT type.  
//     	Coerce special forms like "true", "false" and "optional" to 
//     	their correct types.
//
bool bConvertTermToVariant ( int vt, term_t tArg, _variant_t & vnt )
{
	IP_IDispatch ipId;

	//  Check to see if the current term's atom is a "special" atom
	EKEYWD ekwd = eKeywordCheck( tArg );
	switch ( ekwd )
	{
		case ETRUE:
		case EFALSE:
		{
			vnt.vt = VT_BOOL;
			vnt.boolVal = ekwd == ETRUE 
							? VARIANT_TRUE 
							: VARIANT_FALSE;
			return true;
			break;
		}
		case EOPTIONAL:
		{
			vnt.vt = VT_ERROR;
			vnt.scode = DISP_E_PARAMNOTFOUND;
			return true;
			break;
		}
		//  Not a keyword
		default:
			break;
	}

	//  If the argument is non-specific, convert by term type
	if ( vt == VT_VARIANT )
	{
		//  Get the type of this term
		int iarg = PL_term_type( tArg );
		switch ( iarg )
		{
			//  Integers
			case PL_INTEGER:
				vt = VT_I4;
				break;

			//  Reals
			case PL_FLOAT:
				vt = VT_R8;
				break;

			//  See if this is an interface reference or a list
			case PL_TERM:
			{
				//  See if this is the special SAFEARRAY descriptor functor
				VARTYPE vtTerm = VT_EMPTY;
				term_t tFunctorArg;
				if ( bValidActiveXSafearray( tArg, vtTerm, tFunctorArg ) )
				{
					return bConvertListToVariantSafearray( tFunctorArg, vnt, vtTerm );
				}
				else
				//  See if this is the special VARIANT descriptor functor use for
				//		unusual data types
				if ( bValidActiveXVariant( tArg, vtTerm, tFunctorArg ) )
				{
					return bConvertVariantFunctorToVariant( tFunctorArg, vnt, vtTerm );
				}
				else			
				//  See if this is a garden-variety single-dimensioned, single-element-type list
				if ( PL_is_list( tArg ) )
				{
					return bConvertListToVariantSafearray( tArg, vnt );
				}
				else
				//  See if this is the special IDispatch reference wrapper functor
				if ( bValidActiveXReference( tArg, ipId ) )
				{
					vt = VT_DISPATCH;
					break;
				}
				//  Fall thru to attempt string conversion
			}
			//  Anything vaguely resembling a string
			case PL_ATOM:   
			case PL_STRING:
				vt = VT_BSTR;
				break;
			
			default:
				return false;
		}
	}

	switch ( vt )
	{
		//  Two-byte integers
		case VT_UI2:
	    case VT_I2:
		{
			int i;
			if ( ! PL_get_integer( tArg, & i ) )
				return false;
			short s = i;
			vnt = s;
			break;
		}

		case VT_INT:
		case VT_UI4:
		case VT_UINT:
		case VT_I4:
		{
			int i;
			if ( ! PL_get_integer( tArg, & i ) )
				return false;
			vnt = (long) i;
			break;
		}

		case VT_R4:
		{
			double d;
			if ( ! PL_get_float( tArg, & d ) )
				return false;
			vnt = (float)d;
			break;
		}

		case VT_R8:
		{
			double d;
			if ( ! PL_get_float( tArg, & d ) )
				return false;
			vnt = d;
			break;
		}

		case VT_BSTR:
		{
			char * pch = NULL;
			if ( ! PL_get_chars( tArg, & pch, CVT_ALL | BUF_DISCARDABLE ) )
				return false;

			_bstr_t bstr = pch;
			vnt = bstr;
			break;
		}

		case VT_BOOL:
		{
			char * pch = NULL;
			if ( ! PL_get_chars( tArg, & pch, CVT_ALL | BUF_DISCARDABLE ) )
				return false;
			bool b = ::_strcmpi("true",pch) == 0;
			vnt.Clear();
			new (& vnt) _variant_t( b );
			break;
		}

		case VT_DISPATCH:
		{
			if ( ! ipId )
			{
				if ( ! bValidActiveXReference( tArg, ipId ) )
					return false;
			}
			vnt.Clear();
			new (&vnt) _variant_t( (IDispatch*) ipId );
			break;
		}

		default:
			return false;
	}
	return true;
}


//
//      Using the ITypeInfo interface if necessary, return the
//      required VARTYPE for the TYPEDESC given.  This is normally
//      trivial, but may required dereferencing USERDEFINED types
//      through aliases.  ENUMs are the only supported user-defined types.
//
VARTYPE vtGetVarType ( IP_ITypeInfo & ipInfo, TYPEDESC & tdesc )
{
	VARTYPE vt = tdesc.vt;

	switch ( vt )
	{
		case VT_USERDEFINED:
		{
			HREFTYPE href = tdesc.hreftype;
			IP_ITypeInfo ipTiHref;
			HRESULT hr = ipInfo->GetRefTypeInfo( href, & ipTiHref );
			if ( FAILED(hr) )
				break;
			TYPEATTR * pTypeAttr;
			hr = ipTiHref->GetTypeAttr( & pTypeAttr );
			if ( FAILED(hr) )
				break;
			TYPEKIND tk = pTypeAttr->typekind;
			switch ( tk )
			{
				case TKIND_ALIAS:
					vt = pTypeAttr->tdescAlias.vt;
					if ( vt == VT_USERDEFINED )
						vt = vtGetVarType( ipTiHref, pTypeAttr->tdescAlias );
					break;

				//BUGBUG:  The only user-defined type we allow is ENUMs
				case TKIND_ENUM:
					//  Convert ENUMs to 32-bit ints
					vt = VT_I4;
					break;

				default:
					vt = VT_ILLEGAL;
					break;
			}
			ipTiHref->ReleaseTypeAttr( pTypeAttr );
		}       // Success.. fall thru

		default:
			break;
	}
	return vt;
}
